#ifdef WIN32
#pragma warning(disable:4996)
#endif
#include <iostream>
#include <sstream>
#include <fstream>
#include <filesystem>
#include <thread>
#include <codecvt>
#include <climits>

#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/thread.hpp>
#include <boost/process/search_path.hpp>

#include <fcf/except.hpp>
#include <fcf/system.hpp>
#include <fcf/log.hpp>
#include <fcf/filesystem.hpp>
#include <fcf/net/protocol/http.hpp>

#include "tools.hpp"
#include "configuration.hpp"
#include "endpoint.hpp"

namespace application {

#define MAX_PACK   64*1024
#define STACK_SIZE MAX_PACK*4

namespace {
  std::filesystem::path get_temporary_path(){
    std::filesystem::path path = std::filesystem::temp_directory_path();
    std::stringstream ss;
    ss << "fcf-framework-transfer-data-" << boost::uuids::random_generator()();
    path /= ss.str().c_str();
    return path;
  }

  class file_buffer_t {
  public:
    file_buffer_t()
      : _path(get_temporary_path()), _stream(_path, std::fstream::out | std::fstream::in | std::fstream::app | std::fstream::binary), _first_read(true) {
      
    }

    ~file_buffer_t(){
      try { _stream.close(); } catch (std::exception e) {}
      try { 
        std::filesystem::remove(_path); 
      }
      catch (std::exception e) {
        fcf::err << "ERROR REMOVE FILE TEMP FILE " << _path.string();
      }
    }

    void write(const char* a_datav, size_t a_datac) {
      _stream.write(a_datav, a_datac);
    }

    std::streamsize read(char* a_buffer, size_t a_buffer_size) {
      if (_first_read) {
        _stream.seekg(0, _stream.beg);
        _first_read = false;
      }
      return _stream.readsome(a_buffer, a_buffer_size);
    }

  private:
    std::filesystem::path _path;
    std::fstream          _stream;
    bool                  _first_read;
  };

}

  endpoint_t::endpoint_t(const server_configuration_t& a_server_config, size_t a_handler_index)
    : _slave(a_server_config.slave), 
      _config(a_server_config.endpoints[a_handler_index]), 
      _start_group(0), 
      _stop(false) {

    for(size_t i = 0; i < a_server_config.endpoints.size() && i < a_handler_index; ++i){
      _start_group += (int)a_server_config.endpoints[i].handlers.size();
    }

    if (_config.bind4 != "none")
      _socket4.reset(fcf::net::create_socket(_config.bind4.c_str(), _config.port));
    if (_config.bind6 != "none")
      _socket6.reset(fcf::net::create_socket(_config.bind6.c_str(), _config.port));
  }

  endpoint_t::~endpoint_t(){
    stop();
  }

  void endpoint_t::run() {
    std::unique_lock<std::mutex> lock(_launch_mutex);
    fcf::log << "Running processes...";
    for(size_t hi = 0; hi < _config.handlers.size(); ++hi) {
      handler_configuration_t& hc(_config.handlers[hi]);

      subprocesses_sptr_t spsp(new subprocesses_t(hc.maxMemoryLeakProtection, hc.maxMemory));
      _subprocesses.insert({hi, spsp});

      size_t procindex = 0;
      for(int leak_proteced = 1; leak_proteced >= 0; --leak_proteced) {
        int procc = leak_proteced ? hc.processesLeakProtection: hc.processes;
        for(size_t i = 0; i < procc; ++i) {
          std::vector<std::string> argv1;
          argv1.push_back(hc.script);
          if (_slave)
            argv1.push_back("--slave");
          argv1.push_back("--group");
          argv1.push_back(std::to_string(_start_group + hi));
          argv1.push_back("--server-control-port");
          argv1.push_back(std::to_string(hc.serverControlPort));
          argv1.push_back("--port");
          argv1.push_back(std::to_string(hc.dataPorts[procindex*2]));
          argv1.push_back("--control-port");
          argv1.push_back(std::to_string(hc.controlPorts[procindex*2]));
          argv1.push_back("--host");
          argv1.push_back("localhost");
          if (!!procindex) {
            argv1.push_back("--disable-sys");
            argv1.push_back("--disable-cron");
          }
          if (leak_proteced)
            argv1.push_back("--leak-protected");
            

          std::vector<std::string> argv2;
          argv2.push_back(hc.script);
          if (_slave)
            argv2.push_back("--slave");
          argv2.push_back("--group");
          argv2.push_back(std::to_string(_start_group + hi));
          argv2.push_back("--server-control-port");
          argv2.push_back(std::to_string(hc.serverControlPort));
          argv2.push_back("--port");
          argv2.push_back(std::to_string(hc.dataPorts[procindex*2+1]));
          argv2.push_back("--control-port");
          argv2.push_back(std::to_string(hc.controlPorts[procindex*2+1]));
          argv2.push_back("--host");
          argv2.push_back("localhost");
          if (!!procindex) {
            argv2.push_back("--disable-sys");
            argv2.push_back("--disable-cron");
          }
          if (leak_proteced)
            argv2.push_back("--leak-protected");


          std::vector<std::pair<std::string, std::string> > envv;
          #ifdef WIN32
            std::string nodePath = ".\\node_modules;" + fcf::get_exe_directory() + ";" + fcf::get_exe_directory() + "\\" + "node_modules";
          #else
            std::string nodePath = "./node_modules:" + fcf::get_exe_directory() + ":" + fcf::get_exe_directory() + "/" + "node_modules";
          #endif
          envv.push_back(std::pair<std::string, std::string>("NODE_PATH", nodePath));

          std::wstring_convert<std::codecvt_utf8<wchar_t>> conv;
          std::string node_path = conv.to_bytes(boost::process::search_path("node").wstring());
          _subprocesses[hi]->append(node_path.c_str(), 
                                    argv1,
                                    argv2,
                                    envv,
                                    hc.nodeOptions[procindex],
                                    hc.user.c_str(), 
                                    hc.password.c_str(), 
                                    hc.dataPorts[procindex*2],
                                    hc.dataPorts[procindex*2+1],
                                    hc.controlPorts[procindex*2], 
                                    hc.controlPorts[procindex*2+1], 
                                    hc.startTimeout, 
                                    hc.stopTimeout, 
                                    !!leak_proteced);
          ++procindex;

        }
      }
      _subprocesses[hi]->run();
    }

    if (_socket4.get()){
      fcf::log << "Listening started " << _config.bind4 << " with port " << _config.port;
      _socket4->set_reuse(true);
      _socket4->bind();
      _socket4->listen(_config.listenQueue);
      _endpoint_thread4 = std::thread(&endpoint_t::_endpoint_handler, std::ref(*this), _socket4.get());
    }

    if (_socket6.get()){
      fcf::log << "Listening started " << _config.bind6 << " with port " << _config.port;
      _socket6->set_reuse(true);
      _socket6->bind();
      _socket6->listen(_config.listenQueue);
      _endpoint_thread6 = std::thread(&endpoint_t::_endpoint_handler, std::ref(*this), _socket6.get());
    }
    
  }

  void endpoint_t::stop() {
    std::unique_lock<std::mutex> lock(_launch_mutex);

    if (_stop)
      return;

    _stop = true;

    if (_socket4.get())
      _socket4->close();
    if (_socket6.get())
      _socket6->close();

    if (_endpoint_thread4.joinable())
      _endpoint_thread4.join();
    if (_endpoint_thread6.joinable())
      _endpoint_thread6.join();

    for(size_t hi = 0; hi < _config.handlers.size(); ++hi) {
      if (_subprocesses[hi].get())
        _subprocesses[hi]->stop();
    }
  }

  bool endpoint_t::is_group_exist(int a_group){
    return a_group >= _start_group && a_group < (_start_group + _config.handlers.size());
  }

  void endpoint_t::restart_group(int a_group){
    if (!is_group_exist(a_group))
      return;
    int handler_index = a_group - _start_group;
    _subprocesses[handler_index]->restart();
  }

  void endpoint_t::_endpoint_handler(fcf::net::socket_t* a_socket){
    while(!_stop){
      fcf::net::socket_t* socket = 0;
      try {
        fcf::net::socket_sptr_t socket = fcf::net::socket_sptr_t(a_socket->accept());
        boost::thread::attributes attrs;
        attrs.set_stack_size(STACK_SIZE);
        boost::thread(attrs, boost::bind(&endpoint_t::_handler, this, socket)).detach();
      } catch(std::exception& e){
        if (!_stop)
          fcf::err << "Accept error: " << e.what();
      } catch (...) {
        if (!_stop)
          fcf::err << "Unknown Exception in accept";
      }

    }
  }

  void endpoint_t::_handler(fcf::net::socket_sptr_t a_socket){
    int                                     offset = 0;
    char                                    big_data_buffer[MAX_PACK];
    char                                    first_buffer[MAX_PACK*2];
    int                                     first_buffer_offset = 0;
    std::vector<char>                       residual_buffer;
    unsigned long long                      total_size = 0;
    unsigned long long                      rec_total_size = 0;
    fcf::net::socket_sptr_t                 process_socket;
    fcf::net::protocol::http_header_info_t  request_header_info;
    fcf::net::protocol::http_header_info_t  response_header_info;
    int                                     handler_index = -1;
    int                                     process_index = -1;
    unsigned long long                      full_pack;
    bool                                    first_transaction = true;
    bool                                    completed = false;
    bool                                    is_big_data = false;
    subprocesses_t::subprocesses_state_t*   sps = 0;
    std::shared_ptr<file_buffer_t>          file_buffer;
    int client_socket_timeout;
    int max_request_timeout;
    int server_socket_timeout;
    int max_response_timeout;

    try {

      while(true) {
        std::chrono::time_point start_request = std::chrono::system_clock::now();

        client_socket_timeout = _config.clientSocketTimeout;
        max_request_timeout   = _config.maxRequestTimeout;
        server_socket_timeout = _config.serverSocketTimeout;
        max_response_timeout  = _config.maxResponseTimeout;

        if (first_transaction)
          a_socket->set_timeout(client_socket_timeout);

        bool first_request = true;
        // read header
        bool break_loop = false;
        while(true) {

          completed = false;

          if (!first_transaction && first_request)
            a_socket->set_timeout(_config.keepAliveTimeout);

          int recc = 0;

          try {
            recc = a_socket->recv(&first_buffer[first_buffer_offset], MAX_PACK);
            if (!recc)
              throw std::runtime_error("Failed to get data from client");
          } catch(const fcf::except::socketagain& e){
            if (!first_transaction && first_request){
              break_loop = true;
              break;
            } else {
              throw e;
            }
          }

          if (!first_transaction && first_request)
            a_socket->set_timeout(client_socket_timeout);

          first_request = false;

          if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
            throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());
          
          first_buffer_offset += recc;
          total_size = first_buffer_offset;
          rec_total_size = first_buffer_offset;
          if (!fcf::net::protocol::parse_http_header_request(first_buffer, first_buffer_offset, &request_header_info)){
            if (first_buffer_offset > MAX_PACK)
              throw std::runtime_error("Request header size exceeds maximum packet size");
            continue;
          }
          break;
        }

        if (break_loop)
          break;

        first_transaction = false;

        // find handler configuration
        // width  0 - host and path
        //        1 - host and path (partially) (the 'path_cmp_weigth' variable is taken into account)
        //        2 - host and path (none)
        //        3 - path
        //        4 - path (partially) (the 'path_cmp_weigth' variable is taken into account)
        //        5 - all
        int path_cmp_weigth = 0;
        int width = INT_MAX;
        int handlerc = (int)_config.handlers.size();
        int new_handler_index = -1;
        for (int hi = 0; hi < handlerc; ++hi) {
          handler_configuration_t& hconf(_config.handlers[hi]);
          bool cur_host_empty = hconf.host.empty();
          bool cur_host_iseq = hconf.host == request_header_info.host;

          bool cur_path_empty = hconf.path.empty();
          bool cur_path_iseq = false;
          int  cur_path_cmp_width = 0;
          bool cur_path_match = !cur_path_empty ? cmp_path_template(hconf.path.c_str(), request_header_info.host.c_str(), &cur_path_iseq, &cur_path_cmp_width)
                                                : true;

          if (!cur_host_empty && cur_host_iseq &&
              !cur_path_empty && cur_path_iseq) {
            new_handler_index = hi;
            width = 0;
            path_cmp_weigth = 0;
            break;
          } else if (width >= 1 &&
                    !cur_host_empty && cur_host_iseq && 
                    !cur_path_empty && cur_path_match && 
                    (width != 1 || path_cmp_weigth > cur_path_cmp_width)) {
            new_handler_index = hi;
            width = 1;
            path_cmp_weigth = cur_path_cmp_width;
          } else if (width > 2 &&
                    !cur_host_empty && cur_host_iseq && 
                    cur_path_empty) {
            new_handler_index = hi;
            width = 2;
            path_cmp_weigth = 0;
          } else if (width > 3 &&
                    cur_host_empty && 
                    !cur_path_empty && cur_path_iseq){
            new_handler_index = hi;
            width = 3;
            path_cmp_weigth = 0;
          } else if (width >= 4 &&
                    cur_host_empty && 
                    (width != 4 || path_cmp_weigth > cur_path_cmp_width)){
            new_handler_index = hi;
            width = 4;
            path_cmp_weigth = cur_path_cmp_width;
          } else if (width > 5 &&
                    cur_host_empty && 
                    cur_path_empty){
            new_handler_index = hi;
            width = 5;
            path_cmp_weigth = 0;
          }
        }

        if (new_handler_index == -1)
          throw std::runtime_error(std::string(std::string("Missing request handler for host '") + request_header_info.host + "' with path '" + request_header_info.path + "'").c_str());

        // the definition of a method processing
        request_processing_information_t rpi = _subprocesses[new_handler_index]->get_request_processing_information(request_header_info.path, request_header_info.context);
        if (rpi.client_socket_timeout != -1)
          client_socket_timeout = rpi.client_socket_timeout;
        if (rpi.max_request_timeout != -1)
          max_request_timeout = rpi.max_request_timeout;
        if (rpi.server_socket_timeout != -1)
          server_socket_timeout = rpi.server_socket_timeout;
        if (rpi.max_response_timeout != -1)
          max_response_timeout = rpi.max_response_timeout;

        if (rpi.memory_leak_protection) {

          // reading the residual message
          full_pack   = request_header_info.content_length + request_header_info.header_end + 4;
          is_big_data = full_pack >= 500000;
          if (!is_big_data) {
            if (first_buffer_offset < full_pack)
              residual_buffer.resize(full_pack - first_buffer_offset);
            while (rec_total_size < full_pack) {
              rec_total_size += a_socket->recv(&residual_buffer[rec_total_size - first_buffer_offset], full_pack - rec_total_size < MAX_PACK ? full_pack - rec_total_size : MAX_PACK);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }
          } else if (rec_total_size < full_pack) {
            file_buffer.reset(new file_buffer_t());
            while (rec_total_size < full_pack) {
              size_t size = a_socket->recv(&big_data_buffer[0], full_pack - rec_total_size < MAX_PACK ? full_pack - rec_total_size : MAX_PACK);
              rec_total_size += size;
              file_buffer->write(&big_data_buffer[0], size);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }
          }

          // reconnect to process
          handler_configuration_t& handler_config(_config.handlers[new_handler_index]);

          if (new_handler_index != handler_index) {
            int old_handler_index = handler_index;
            handler_index = new_handler_index;

            if (process_index != -1){
              _subprocesses[old_handler_index]->release_process(process_index);
              process_index = -1;
            }

            process_index = _subprocesses[handler_index]->get_process(SLP_YES);

            sps = &(_subprocesses[handler_index]->get_process_info(process_index));
            process_socket.reset(fcf::net::create_socket("127.0.0.1", sps->port));
            process_socket->set_timeout(server_socket_timeout);
            process_socket->connect();
          }

          // send header with data
          offset = 0;
          while(offset < first_buffer_offset){
            offset += process_socket->send(&first_buffer[offset], first_buffer_offset - offset);
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }

          // send residual message to process
          offset = 0;
          if (!is_big_data) {
            //residual_buffer
            while((offset + first_buffer_offset) < full_pack) {
              offset += process_socket->send(&residual_buffer[offset], (full_pack - first_buffer_offset) - offset);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }
          } else if (first_buffer_offset < full_pack) {
            while((offset + first_buffer_offset) < full_pack) {
              size_t pack_size = (full_pack - first_buffer_offset) - offset < MAX_PACK ? (full_pack - first_buffer_offset) - offset : MAX_PACK;
              file_buffer->read(&big_data_buffer[0], pack_size);

              size_t part_size = 0;
              while(part_size < pack_size) {
                part_size += process_socket->send(&big_data_buffer[part_size], pack_size - part_size);
                if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                  throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
              }
              offset += pack_size;
            }
          }

          // get response
          std::chrono::time_point start_response = std::chrono::system_clock::now();
          offset = 0;
          bool first = true;
          process_socket->set_timeout(max_response_timeout);
          while(true) {
            int recc = process_socket->recv(&first_buffer[offset], MAX_PACK);
            if (!recc)
              throw std::runtime_error("Failed to get data from server");

            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
              throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());

            if (first)
              process_socket->set_timeout(server_socket_timeout);

            first = false;
            offset += recc;
            if (!fcf::net::protocol::parse_http_header_response(first_buffer, offset, &response_header_info)){
              if (offset > MAX_PACK)
                throw std::runtime_error("Response header size exceeds maximum packet size");
              continue;
            }
            total_size = offset;
            break;
          }

          // send header with data to client
          full_pack = response_header_info.content_length + response_header_info.header_end + 4;
          offset = 0; 
          while (offset < total_size) {
            if (!sps->active) {
              _buffered_forwarding(&first_buffer[offset],
                                   total_size - offset,
                                   full_pack - offset,
                                   a_socket,
                                   process_socket,
                                   handler_index,
                                   process_index,
                                   start_response,
                                   request_header_info,
                                   max_response_timeout);
              completed = true;
              break;
            }
            offset += a_socket->send(&first_buffer[offset], total_size - offset > MAX_PACK ? MAX_PACK : (unsigned int)(total_size - offset));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
              throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }

          // resend data to client
          if (!completed) {
            while(total_size < full_pack) {
              if (!sps->active) {
                _buffered_forwarding(&first_buffer[0],
                                     0,
                                     full_pack - total_size,
                                     a_socket,
                                     process_socket,
                                     handler_index,
                                     process_index,
                                     start_response,
                                     request_header_info,
                                     max_response_timeout);
                completed = true;
                break;
              }

              offset = process_socket->recv(&first_buffer[0], MAX_PACK);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
                throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());

              if (!sps->active) {
                _buffered_forwarding(&first_buffer[0],
                                     offset,
                                     full_pack - total_size,
                                     a_socket,
                                     process_socket,
                                     handler_index,
                                     process_index,
                                     start_response,
                                     request_header_info,
                                     max_response_timeout);
                completed = true;
                break;
              }

              int sendc = 0;
              while(sendc < offset){
                sendc += a_socket->send(&first_buffer[sendc], offset - sendc);
                if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
                  throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
              }

              total_size += offset;
            }
          }
        } else {

          // reconnect to process
          handler_configuration_t& handler_config(_config.handlers[new_handler_index]);

          if (new_handler_index != handler_index) {
            int old_handler_index = handler_index;
            handler_index = new_handler_index;

            if (process_index != -1){
              _subprocesses[old_handler_index]->release_process(process_index);
              process_index = -1;
            }

            process_index = _subprocesses[handler_index]->get_process(SLP_NO);
            sps = &(_subprocesses[handler_index]->get_process_info(process_index));
            process_socket.reset(fcf::net::create_socket("127.0.0.1", sps->port));
            process_socket->set_timeout(server_socket_timeout);
            process_socket->connect();
          }

          // send header with data
          offset = 0;
          while(offset < total_size){
            offset += process_socket->send(&first_buffer[offset], total_size - offset > MAX_PACK ? MAX_PACK : (unsigned int)(total_size - offset));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }

          // resend data to process
          full_pack = request_header_info.content_length + request_header_info.header_end + 4;
          while(total_size < full_pack){
            offset = a_socket->recv(&first_buffer[0], MAX_PACK);
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
              throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());

            int sendc = 0;
            while(sendc < offset){
              sendc += process_socket->send(&first_buffer[sendc], offset - sendc);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_request_timeout)
                throw fcf::except::max_request_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }

            total_size += offset;
          }

          // get response
          std::chrono::time_point start_response = std::chrono::system_clock::now();
          offset = 0;
          bool first = true;
          process_socket->set_timeout(max_response_timeout);
          while(true) {
            int recc = process_socket->recv(&first_buffer[offset], MAX_PACK);
            if (!recc)
              throw std::runtime_error("Failed to get data from server");

            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
              throw fcf::except::max_request_timeout(a_socket->get_address(), a_socket->get_port());

            if (first)
              process_socket->set_timeout(server_socket_timeout);

            first = false;
            offset += recc;
            if (!fcf::net::protocol::parse_http_header_response(first_buffer, offset, &response_header_info)){
              if (offset > MAX_PACK)
                throw std::runtime_error("Response header size exceeds maximum packet size");
              continue;
            }
            total_size = offset;
            break;
          }

          // send header with data to client
          offset = 0; 
          while(offset < total_size){
            offset += a_socket->send(&first_buffer[offset], total_size - offset > MAX_PACK ? MAX_PACK : (unsigned int)(total_size - offset));
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_response).count() > max_response_timeout)
              throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
          }

          // resend data to client
          full_pack = response_header_info.content_length + response_header_info.header_end + 4;
          while(total_size < full_pack){
            offset = process_socket->recv(&first_buffer[0], MAX_PACK);
            if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_response_timeout)
              throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());

            int sendc = 0;
            while(sendc < offset){
              sendc += a_socket->send(&first_buffer[sendc], offset - sendc);
              if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start_request).count() > max_response_timeout)
                throw fcf::except::max_response_timeout_fi(a_socket->get_address(), a_socket->get_port(), request_header_info.host.c_str(), request_header_info.path.c_str());
            }

            total_size += offset;
          }

        }

        if (response_header_info.connection != "keep-alive")
          break;

        if (process_index != -1 && !_subprocesses[handler_index]->is_active(process_index))
          break;

        if (_stop)
          break;

        offset              = 0;
        first_buffer_offset = 0;
        total_size          = 0;
        completed           = false;
      }

    } catch(const std::exception& e) {
      fcf::trc  << "IP:" << a_socket->get_address() << " PORT:" << a_socket->get_port() << " " << e.what();
    } catch (...) {
      fcf::trc << "Unknown Exception in handler";
    }

    if (process_index != -1){
      _subprocesses[handler_index]->release_process(process_index);
      process_index = -1;
    }

  }

  template<typename TimePoint>
  void endpoint_t::_buffered_forwarding(const char*                             a_unsent_data_buffer,
                                        size_t                                  a_unsent_data_buffer_size,
                                        unsigned long long                      a_total_unset_data_size,
                                        fcf::net::socket_sptr_t                 a_client_socket,
                                        fcf::net::socket_sptr_t                 a_process_socket,
                                        int                                     a_handler_index,
                                        int&                                    a_process_index,
                                        TimePoint                               a_start_response,
                                        fcf::net::protocol::http_header_info_t& a_request_header_info,
                                        int                                     a_max_response_timeout) {
    std::vector<char> buffer(a_total_unset_data_size - a_unsent_data_buffer_size);
    unsigned long long offset = 0;

    while (offset < (a_total_unset_data_size - a_unsent_data_buffer_size)) {
      unsigned long long residual = (a_total_unset_data_size - a_unsent_data_buffer_size) - offset;
      offset += a_process_socket->recv(&buffer[offset], residual < MAX_PACK ? residual : MAX_PACK);
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }

    if (a_process_index != -1 && a_handler_index != -1){
      _subprocesses[a_handler_index]->release_process(a_process_index);
      a_process_index = -1;
    }

    offset = 0;
    while (offset < a_unsent_data_buffer_size) {
      offset += a_client_socket->send(&a_unsent_data_buffer[offset], a_unsent_data_buffer_size - offset < MAX_PACK ? a_unsent_data_buffer_size - offset : MAX_PACK);
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }

    while (offset < a_total_unset_data_size) {
      offset += a_client_socket->send(&buffer[offset - a_unsent_data_buffer_size], a_total_unset_data_size - offset < MAX_PACK ? a_total_unset_data_size - offset : MAX_PACK);
      if (std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - a_start_response).count() > a_max_response_timeout)
        throw fcf::except::max_response_timeout_fi(a_client_socket->get_address(), a_client_socket->get_port(), a_request_header_info.host.c_str(), a_request_header_info.path.c_str());
    }
  }


}
